﻿ThreatLogger = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "Parser-3.0")

local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local reportFrame, reportFrameMaster, playerDamage, lastPlayerHit

function ThreatLogger:OnInitialize()
	reportFrame = _G.ThreatLoggerEditBox
	reportFrameMaster = _G.ThreatLoggerEditFrame
	reportFrame:SetParent(reportFrameMaster)
	reportFrameMaster:SetScrollChild(reportFrame)
	-- reportFrame:SetMultiLine(true)
	reportFrameMaster:SetWidth(500)
	reportFrameMaster:SetHeight(500)
	reportFrame:SetWidth(480)
	reportFrame:SetHeight(500)
	reportFrame:SetTextInsets(10, 10, 10, 10)
	reportFrame:SetFont("FRIZQT__.TTF",8)
	closeButton = CreateFrame("Button", "$parentClose", reportFrameMaster, "UIPanelCloseButton")
	closeButton:SetPoint("TOPRIGHT", reportFrameMaster, "TOPRIGHT")
	tinsert(UISpecialFrames, "ThreatLoggerEditFrame");

	reportFrameMaster:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4}
	})
	-- reportFrame:SetAutoFocus(false)
	reportFrame:EnableMouse(true)
	reportFrameMaster:SetBackdropColor(0,0,0,0.6)
	reportFrameMaster:SetPoint("CENTER", UIParent, "CENTER")
	reportFrameMaster:Hide()
end

local logs = {}
local logTimes = {}
local print = false
function ThreatLogger:StartLogging(lPrint)
	for i = 1, #logs do
		tremove(logs)
		tremove(logTimes)
	end
	
	self:RegisterParserEvent({
		eventType = 'Damage'
	}, "parseDamage")
	self:RegisterParserEvent({
		eventType = 'Heal',
		sourceID_not = false,
		recipientID_not = false
	}, "parseHeal")
	self:RegisterParserEvent({
		eventType = 'Gain',
		sourceID_not = false,
		recipientID_not = false
	}, "parseGain")
	self:RegisterParserEvent({
		eventType = 'Cast'
	}, "parseCast")	
	self:RegisterEvent("UNIT_TARGET")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	print = lPrint
end

function ThreatLogger:StopLogging()
	self:UnregisterAllParserEvents()
	self:UnregisterAllEvents()
end

function ThreatLogger:UNIT_TARGET(arg1)
	if arg1 then
		tinsert(logs, string.format("%s > target > %s", UnitName(arg1), UnitName(arg1 .. "target") or "<none>")) 
	else
		tinsert(logs, string.format("%s > target > %s", "<unknown>", "<unknown>")) 
	end
	tinsert(logTimes, GetTime())
	
	if UnitIsUnit(arg1, "target") then
		tinsert(logs, string.format("Player did %s damage since last ToT change, last hit was %s, threat range is %2.2f-%2.2f", playerDamage, lastPlayerHit, (playerDamage-lastPlayerHit) / 1.1, playerDamage/1.1))
		tinsert(logTimes, GetTime())
	end
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

function ThreatLogger:UNIT_SPELLCAST_SUCCEEDED(arg1, arg2, arg3, arg4)
	tinsert(logs, string.format("%s > cast %s > %s", UnitName(arg1) or arg1, arg2, arg3)) 
	tinsert(logTimes, GetTime())
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

function ThreatLogger:Report()
	local s = ""
	for i = 1, #logs do
		s = s .. string.format("%s: %s\n", logTimes[i], logs[i])
	end
	s = s .. string.format("Player did %s damage since last reset, last hit was %s, threat range is %2.2f-%2.2f", playerDamage, lastPlayerHit, (playerDamage-lastPlayerHit) / 1.1, playerDamage/1.1)
	--_G[reportFrame:GetName() .. "Edit"]:SetText(s)
	reportFrame:SetText(s)
	ThreatLoggerEditFrame:UpdateScrollChildRect()
	ThreatLoggerEditFrame:Show()
end

function ThreatLogger:parseDamage(info)
	if not info.sourceID and not info.recipientID then return end
	tinsert(logs, string.format("%s > %s damage (%s) > %s", info.sourceName, info.amount, info.damageType, info.recipientName)) 
	tinsert(logTimes, GetTime())
	if info.sourceID and UnitIsUnit(info.sourceID, "player") then
		lastPlayerHit = info.amount
		playerDamage = playerDamage + info.amount
	end
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

function ThreatLogger:parseHeal(info)
	local amount = min(info.amount, UnitHealthMax(info.recipientID) - UnitHealth(info.recipientID))
	tinsert(logs, string.format("%s > %s heal > %s", info.sourceName, amount, info.recipientName)) 
	tinsert(logTimes, GetTime())
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

function ThreatLogger:parseGain(info)
	local amount = max(info.amount, UnitManaMax(info.recipientID) - UnitMana(info.recipientID))
	tinsert(logs, string.format("%s > %s gain %s > %s", info.sourceName, amount, info.attribute, info.recipientName)) 
	tinsert(logTimes, GetTime())
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

function ThreatLogger:parseCast(info)
	local bap, pap, map, ap = -1, -1, -1, -1
	if info.sourceID then
		bap, pap, map = UnitAttackPower(info.sourceID)
		ap = bap + pap - map
	end
	tinsert(logs, string.format("%s > Cast %s > %s [AP: %s (%s + %s - %s)]", info.sourceName, info.abilityName, info.recipientName or "<unknown>", ap, bap, pap, map)) 
	tinsert(logTimes, GetTime())
	if print then self:Print("%s: %s", logTimes[#logTimes], logs[#logs]) end
end

local options = {
	type = "group",
	args = {
		start = {
			type = "execute",
			name = "Start",
			desc = "Start",
			func = function() 
				playerDamage = 0
				lastPlayerHit = 0
				ThreatLogger:StartLogging(false)
				ThreatLogger:Print("Logging started (display off).")
			end
		},
		start2 = {
			type = "execute",
			name = "Start With Display",
			desc = "Start With Display",
			func = function() 
				playerDamage = 0
				lastPlayerHit = 0
				ThreatLogger:StartLogging(true)
				ThreatLogger:Print("Logging started (display on).")
			end
		},		
		stop = {
			type = "execute",
			name = "Stop",
			desc = "Stop",
			func = function()
				ThreatLogger:StopLogging()
				ThreatLogger:Print("Logging stopped.")
			end
		},		
		report = {
			type = "execute",
			name = "Report",
			desc = "Report",
			func = function()
				ThreatLogger:Report()
			end
		},
		reset = {
			type = "execute",
			name = "Reset",
			desc = "Reset",
			func = function()
				ThreatLogger:StopLogging()
				playerDamage = 0
				lastPlayerHit = 0
				ThreatLogger:StartLogging(print)
				ThreatLogger:Print("Logs reset")
			end
		}
	}
}
ThreatLogger:RegisterChatCommand({"/tlog"}, options, "THREATLOGGER")